#include <cassert>
#include <cstdint>
#include <climits>
using std::numeric_limits;

#include <algorithm>
using std::generate;

#include "bitconvert.h"

//const size_t N = sizeof(uint32_t);
//const size_t twoN = 2*N;

// x = (a*x + c) mod m
// x = (a*x + c) mod 2^32
// a = 1664525, c = 1013904223
// a = 22695477, c = 1
// a = 1103515245, c = 12345
// a = 134775813, c = 1
// a = 214013, c = 2531011
template <uint32_t a, uint32_t c, uint32_t m = 4294967295U>
class linear_congruential
{
private:
	uint32_t x;
public:	
	static const bool is_wordsize = ( m == 4294967295U );
	linear_congruential() : x(0) {};
	linear_congruential(uint32_t x_) : x(x_) {};
	void seed(uint32_t x_) { x = x_; }
	uint32_t operator()() 
	{ 
		return x = (is_wordsize ? a*x+c : (a*x+c)%m );
	}
};

// x = (a*x^2 + b*x + c) mod 2^32
template <uint32_t a, uint32_t b, uint32_t c>
class quadratic_congruential
{
public:
	static const bool is_wordsize = true;   
	// (!)
};

// j = 7, k = 10
// j = 5, k = 17
// j = 24, k = 55
// j = 65, k = 71
// j = 128, k = 159
// j = 6, k = 31
// j = 31, k = 63
// j = 97, k = 127
// j = 353, k = 521
// j = 168, k = 521
// j = 334, k = 607
// j = 273, k = 607
// j = 418, k = 1279
template <size_t J, size_t K, class G>
class lagged_fibonacci
{
public:
	static const bool is_wordsize = true;
	// (!)
};

class middle16_bits
{
private:
	uint32_t x;
public:
	static const bool is_wordsize = true;
	middle16_bits() : x(0) {}
	middle16_bits(uint32_t x_) : x(x_) {}
	void seed(uint32_t x_) { x = x_; }
	uint32_t operator()() 
	{
		// (!)
		return 0;
	}
};

class middle32_bits
{
private:
	uint32_t x;
public:
	static const bool is_wordsize = true;
	middle32_bits() : x(0) {}
	middle32_bits(uint32_t x_) : x(x_) {}
	void seed(uint32_t x_) { x = x_; }
	uint32_t operator()()
	{ 
		// (!)
		return 0;
	}
};

template <class G>
class rnd
{
private:
	G generator;
	static_assert(G::is_wordsize, "Must be 32 bits wide"); // make sure that G uses m = 2^32
public:
	rnd() : generator() {}
	rnd(uint32_t s) : generator(s) {}
	void seed(uint32_t s) { generator.seed(s); }
	double operator()() { return rand32(); }
	float rand23() // 23-bit fp fraction in [0,1)
	{
		convert32_t C;
		C.I32 = (generator() >> 9) | 0x3F800000;
		return C.F32-1.0;
	}
	double rand52() // 52-bit fp fraction in [0,1)
	{
		// (!)
		return 0.0;
	}
	double rand32() // 32-bit fp fraction in [0,1)
	{
		return generator() * (1.0/4294967296.);
	}
	double rand32Inc01() // 32-bit fp fraction in [0,1]
	{
		return generator() * (1.0/4294967295.);
	}
	double rand32Exc01() // 32-bit fp fraction in (0,1)
	{
		return ( double(generator()) + 0.5 ) * (1.0/4294967296.0); 
	}
	uint32_t randInt(void)
	{
		return generator();
	}
	uint32_t randInt(const uint32_t &n)
	{
		return uint32_t(rand32()*(1+n));
	}
	uint32_t randInt(const uint32_t &n1, const uint32_t &n2) { return n1 + randInt(n2-n1); }
};
